/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.aggregation.transformation;

import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

import org.springframework.stereotype.Service;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.IntNode;
import com.fasterxml.jackson.databind.node.TextNode;
import com.schibsted.spt.data.jslt.Expression;
import com.schibsted.spt.data.jslt.Function;
import com.schibsted.spt.data.jslt.filters.DefaultJsonFilter;
import com.schibsted.spt.data.jslt.impl.AbstractNode;
import com.schibsted.spt.data.jslt.impl.ArraySlicer;
import com.schibsted.spt.data.jslt.impl.DotExpression;
import com.schibsted.spt.data.jslt.impl.ExpressionImpl;
import com.schibsted.spt.data.jslt.impl.ExpressionNode;
import com.schibsted.spt.data.jslt.impl.LetExpression;
import com.schibsted.spt.data.jslt.impl.LiteralExpression;
import com.schibsted.spt.data.jslt.impl.ObjectExpression;
import com.schibsted.spt.data.jslt.impl.PairExpression;

import de.ozgcloud.aggregation.transformation.AggregationMapping.FieldMapping;
import lombok.RequiredArgsConstructor;

/**
 * JSTL implementation of the {@link TransformationService}. This service
 * internally uses JSLT {@link Expression} for {@link Transformation}s
 */
@Service
@RequiredArgsConstructor
public class JSLTransformationService implements TransformationService {

	private static final Pattern LIST_ELEMENT_PATTERN = Pattern.compile("(.+)\\[(\\d+)\\]");
	static final LetExpression[] EMPTY_TEMPLATES = new LetExpression[] {};
	static final Map<String, Function> EMPTY_FUNCTIONS = Map.of();

	private final ObjectMapper objectMapper;
	private final VorgangMapper vorgangMapper;

	@Override
	public Transformation load(String identifier, AggregationMapping mapping) throws TransformationException {
		return JSLTransformation.builder()
				.vorgangMapper(vorgangMapper)
				.objectMapper(objectMapper)
				.identifier(transformToAbstractNode(identifier))
				.script(createExpression(transformToMap(getFieldMappings(mapping))))
				.build();
	}

	private List<FieldMapping> getFieldMappings(AggregationMapping mapping) {
		return Optional.ofNullable(mapping).map(AggregationMapping::getFieldMappings).orElse(Collections.emptyList());
	}

	Map<String, String> transformToMap(List<FieldMapping> mappings) {
		return mappings.stream().collect(Collectors.toMap(FieldMapping::getTargetPath, FieldMapping::getSourcePath));

	}

	Expression createExpression(Map<String, String> mapping) {
		var pairExpressions = mapping.entrySet().stream().map(this::toPairExpression).filter(Optional::isPresent).map(Optional::get)
				.toArray(PairExpression[]::new);
		return new ExpressionImpl(EMPTY_TEMPLATES, EMPTY_FUNCTIONS, buildObjectExpression(pairExpressions));
	}

	Optional<PairExpression> toPairExpression(Map.Entry<String, String> mapping) {
		return transformToAbstractNode(mapping.getValue()).map(path -> buildPairExpression(mapping.getKey(), path));
	}

	ObjectExpression buildObjectExpression(PairExpression... fields) {
		return new ObjectExpression(EMPTY_TEMPLATES, fields, null, null, new DefaultJsonFilter());
	}

	PairExpression buildPairExpression(String key, ExpressionNode expression) {
		return new PairExpression(new LiteralExpression(new TextNode(key), null), expression, null);
	}

	Optional<AbstractNode> transformToAbstractNode(String path) {
		if (path == null) {
			return Optional.empty();
		} else {
			var fields = path.split("\\.");
			AbstractNode current = null;
			for (var field : fields) {
				var listElementMatcher = LIST_ELEMENT_PATTERN.matcher(field);
				if (listElementMatcher.matches()) {
					current = getNodeForListElement(current, listElementMatcher);
				} else {
					current = new DotExpression(field, current, null);
				}
			}
			return Optional.of(current);
		}
	}

	private AbstractNode getNodeForListElement(AbstractNode current, Matcher arrayMatcher) {
		var index = Integer.parseInt(arrayMatcher.group(2));
		var name = arrayMatcher.group(1);
		return slicer(current, name, index);
	}

	private ArraySlicer slicer(AbstractNode parent, String field, int index) {
		var idx = new IntNode(index);
		var node = new DotExpression(field, parent, null);
		return new ArraySlicer(new LiteralExpression(idx, null), false, null, node, null);
	}

}
