/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.aggregation;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.temporal.ChronoUnit;
import java.util.Collections;
import java.util.List;
import java.util.stream.IntStream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.TestConfiguration;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Primary;
import org.springframework.test.context.TestPropertySource;
import org.springframework.test.context.bean.override.mockito.MockitoSpyBean;

import de.ozgcloud.aggregation.warehouse.DocumentEntry;
import de.ozgcloud.aggregation.warehouse.DocumentEntryTestFactory;
import de.ozgcloud.aggregation.warehouse.ObjectNodeTestFactory;
import de.ozgcloud.aggregation.warehouse.WarehouseRepository;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgang;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangHeader;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangHeaderTestFactory;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangId;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangService;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangStub;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangStubTestFactory;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangTestFactory;
import de.ozgcloud.apilib.vorgang.Page;
import de.ozgcloud.common.test.DataITCase;

@DataITCase
@TestPropertySource(properties = { "ozgcloud.command.line.runner.enabled=true" })
class AggregationManagerRunnerITCase {

	@Autowired
	private WarehouseRepository repository;

	@MockitoSpyBean
	private AggregationManagerRunner runner;

	@Test
	void shouldRunAutomatically() {
		verify(runner).run();
	}

	@Nested
	class TestRun {

		@BeforeEach
		void setUp() {
			repository.deleteAll();
		}

		@Test
		void shouldProcessAllData() {
			runner.run();

			assertThat(repository.count()).isEqualTo(TestConfig.NUMBER_OF_VORGAENGE + TestConfig.NUMBER_OF_VORGANG_STUBS);
		}

		@Nested
		class TestVorgangDocument {

			private static final OzgCloudVorgang VORGANG = TestConfig.VORGANGS.getFirst();

			@Test
			void shouldHaveEingangsdatum() {
				runner.run();

				var document = findVorgangDocument();
				assertThat(document.getEingangsdatum()).isCloseTo(VORGANG.getHeader().getCreatedAt(), within(1, ChronoUnit.MILLIS));
			}

			@Test
			void shouldHaveVorgangName() {
				runner.run();

				var document = findVorgangDocument();
				assertThat(document.getVorgangsname()).isEqualTo(VORGANG.getVorgangName());
			}

			@Test
			void shouldHaveStatus() {
				runner.run();

				var document = findVorgangDocument();
				assertThat(document.getStatus()).isEqualTo(VORGANG.getHeader().getStatus().toString());
			}

			@Test
			void shouldNotHaveAnyPayload() {
				runner.run();

				var document = findVorgangDocument();
				assertThat(document.getPayload()).isEmpty();
			}

			private DocumentEntry findVorgangDocument() {
				return repository.findById(VORGANG.getId().toString()).get();
			}
		}

		@Nested
		class TestVorgangStubDocument {

			private static final OzgCloudVorgangStub VORGANG_STUB = TestConfig.VORGANG_STUBS.getFirst();

			@Test
			void shouldHaveValues() {
				var expected = DocumentEntryTestFactory.createBuilder()
						.id(VORGANG_STUB.getId().toString())
						.eingangsdatum(DocumentEntryTestFactory.EINGANGSDATUM.withZoneSameInstant(ZoneId.systemDefault()))
						.payload(ObjectNodeTestFactory.create().removeAll())
						.build();

				runner.run();

				assertThat(findVorgangStubDocument()).usingRecursiveComparison().isEqualTo(expected);

			}

			private DocumentEntry findVorgangStubDocument() {
				return repository.findById(VORGANG_STUB.getId().toString()).get();
			}
		}
	}

	@TestConfiguration
	public static class TestConfig {

		private static final int NUMBER_OF_VORGAENGE = 10;
		public static final List<OzgCloudVorgang> VORGANGS = createVorgaenge();

		private static final int NUMBER_OF_VORGANG_STUBS = 5;
		public static final List<OzgCloudVorgangStub> VORGANG_STUBS = createVorgangStubs();

		private static List<OzgCloudVorgang> createVorgaenge() {
			return IntStream.range(0, NUMBER_OF_VORGAENGE)
					.mapToObj(TestConfig::createVorgang)
					.toList();
		}

		private static OzgCloudVorgang createVorgang(int i) {
			return OzgCloudVorgangTestFactory.createBuilder()
					.id(OzgCloudVorgangId.from("id" + i))
					.vorgangName("test-" + i)
					.vorgangNummer("nummer-" + i)
					.header(createHeader(i))
					.build();
		}

		private static OzgCloudVorgangHeader createHeader(int i) {
			return OzgCloudVorgangHeaderTestFactory.createBuilder()
					.createdAt(ZonedDateTime.now().minusDays(1))
					.aktenzeichen("akt-" + i).build();
		}

		private static OzgCloudVorgangStub mapToStub(OzgCloudVorgang vorgang) {
			return OzgCloudVorgangStub.builder().id(vorgang.getId()).vorgangName(vorgang.getVorgangName())
					.vorgangNummer(vorgang.getVorgangNummer()).header(vorgang.getHeader()).build();
		}

		private static List<OzgCloudVorgangStub> createVorgangStubs() {
			return IntStream.range(0, NUMBER_OF_VORGANG_STUBS).mapToObj(TestConfig::createVorgangStub).toList();
		}

		private static OzgCloudVorgangStub createVorgangStub(int i) {
			return OzgCloudVorgangStubTestFactory.createBuilder()
					.id(OzgCloudVorgangId.from("stub-id" + i))
					.build();
		}

		@Bean
		@Primary
		OzgCloudVorgangService vorgangService() {
			final List<OzgCloudVorgangStub> stubs = VORGANGS.stream().map(TestConfig::mapToStub).toList();
			final OzgCloudVorgangService vorgangService = mock(OzgCloudVorgangService.class);
			when(vorgangService.find(any(), any())).thenAnswer(i -> {
				final Page page = i.getArgument(1, Page.class);
				if (page.getOffset() >= stubs.size()) {
					return Collections.emptyList();
				} else {
					final List<OzgCloudVorgangStub> subList = stubs.subList(page.getOffset(),
							Math.min(stubs.size(), page.getLimit() + page.getOffset()));
					return subList;
				}
			});

			when(vorgangService.getById(any()))
					.thenAnswer(i -> VORGANGS.stream().filter(vorgang -> vorgang.getId().equals(i.getArgument(0)))
							.findFirst().orElse(null));

			when(vorgangService.findDeleted()).thenAnswer(i -> VORGANG_STUBS.stream());

			return vorgangService;
		}

	}

}
