/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.aggregation;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.List;

import org.apache.commons.lang3.RandomUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentMatcher;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.aggregation.AggregationManagerRunner.Execution;
import de.ozgcloud.aggregation.transformation.AggregationMapping;
import de.ozgcloud.aggregation.transformation.AggregationMappingTestFactory;
import de.ozgcloud.aggregation.transformation.Transformation;
import de.ozgcloud.aggregation.transformation.TransformationService;
import de.ozgcloud.aggregation.transformation.VorgangMapper;
import de.ozgcloud.aggregation.warehouse.WarehouseRepository;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangService;

class AggregationManagerRunnerTest {

	@Mock
	private OzgCloudVorgangService vorgangService;
	@Mock
	private AggregationManagerConfiguration config;
	@Mock
	private TransformationProperties transformationProperties;
	@Mock
	private TransformationService transformationService;
	@Mock
	private WarehouseRepository repository;
	@Mock
	private VorgangMapper vorgangMapper;
	@Mock
	private static Aggregator aggregator;
	@Spy
	@InjectMocks
	private AggregationManagerRunnerImpl runner;

	@Nested
	class TestRun {
		private final String identifier = LoremIpsum.getInstance().getWords(1);

		@BeforeEach
		void mock() {
			when(transformationProperties.getIdentifier()).thenReturn(identifier);
			doNothing().when(runner).runWithTransformation(any(), any());
		}

		@Test
		void shouldGetAggregationMappings() {
			runner.run();

			verify(transformationProperties).getAggregationMappings();
		}

		@Test
		void shouldGetIdentifier() {
			runner.run();

			verify(transformationProperties).getIdentifier();
		}

		@Nested
		class TestOnAggregationMappingConfigured {
			private final List<AggregationMapping> aggregationMappings = List.of(AggregationMappingTestFactory.create(),
					AggregationMappingTestFactory.create());

			@Mock
			private Transformation firstTransformation;
			@Mock
			private Transformation secondTransformation;

			@BeforeEach
			void mock() {
				when(transformationProperties.getAggregationMappings()).thenReturn(aggregationMappings);
				when(transformationService.load(any(), any())).thenReturn(firstTransformation, secondTransformation);
			}

			@Test
			void shouldLoadTransformationForEachAggregationMapping() {
				runner.run();

				aggregationMappings.forEach(mapping -> verify(transformationService).load(identifier, mapping));
			}

			@Test
			void shouldRunWithTransformationForEachTransformation() {
				runner.run();

				verify(runner).runWithTransformation(firstTransformation, aggregationMappings.get(0));
				verify(runner).runWithTransformation(secondTransformation, aggregationMappings.get(1));
			}
		}

		@Nested
		class TestOnAggregationMappingsNull {
			@Mock
			private Transformation transformation;

			@BeforeEach
			void mock() {
				when(transformationProperties.getAggregationMappings()).thenReturn(null);
				when(transformationService.load(any(), any())).thenReturn(transformation);
			}

			@Test
			void shouldLoadTransformation() {
				runner.run();

				verify(transformationService).load(identifier, null);
			}

			@Test
			void shouldCallRunWithDefaultTransformation() {
				runner.run();

				verify(runner).runWithDefaultTransformation(transformation);
			}
		}

		@Nested
		class TestOnAggregationMappingsEmpty {
			@Mock
			private Transformation transformation;

			@BeforeEach
			void mock() {
				when(transformationProperties.getAggregationMappings()).thenReturn(Collections.emptyList());
				when(transformationService.load(any(), any())).thenReturn(transformation);
			}

			@Test
			void shouldLoadTransformation() {
				runner.run();

				verify(transformationService).load(identifier, null);
			}

			@Test
			void shouldCallRunWithDefaultTransformation() {
				runner.run();

				verify(runner).runWithDefaultTransformation(transformation);
			}
		}
	}

	@Nested
	class TestRunWithDefaultTransformation {

		@Mock
		private Transformation transformation;

		@Test
		void shouldCallRunWithTransformation() {
			doNothing().when(runner).runWithTransformation(any(), any());

			runner.runWithDefaultTransformation(transformation);

			verify(runner).runWithTransformation(eq(transformation), isNull());
		}
	}

	@Nested
	class TestRunWithTransformation {

		@Mock
		private Transformation transformation;
		private final AggregationMapping aggregationMapping = AggregationMappingTestFactory.create();
		private final ArgumentMatcher<Execution> hasTransformation = execution -> execution.getTransformation().equals(transformation);

		@BeforeEach
		void mock() {
			doReturn(aggregator).when(runner).prepareAggregator(any(), any());
		}

		@Test
		void shouldCallPrepareAggregator() {
			runWithTransformation();

			verify(runner).prepareAggregator(argThat(hasTransformation), eq(aggregationMapping));
		}

		@Test
		void shouldLoadAggregation() {
			runWithTransformation();

			verify(aggregator).aggregate();
		}

		private void runWithTransformation() {
			runner.runWithTransformation(transformation, aggregationMapping);
		}
	}

	@Nested
	class TestPrepareAggregator {

		@Mock
		private Execution execution;
		private final AggregationMapping aggregationMapping = AggregationMappingTestFactory.create();
		private final int batchSize = RandomUtils.insecure().randomInt();

		@BeforeEach
		void mock() {
			when(config.getFetchingBatchSize()).thenReturn(batchSize);
			when(aggregator.withExecution(any())).thenReturn(aggregator);
			when(aggregator.withAggregationMapping(any())).thenReturn(aggregator);
			when(aggregator.withBatchSize(anyInt())).thenReturn(aggregator);
		}

		@Test
		void shouldGetBatchSize() {
			runner.prepareAggregator(execution, aggregationMapping);

			verify(config).getFetchingBatchSize();
		}

		@Test
		void shouldSetExecution() {
			runner.prepareAggregator(execution, aggregationMapping);

			verify(aggregator).withExecution(execution);
		}

		@Test
		void shouldSetAggregationMapping() {
			runner.prepareAggregator(execution, aggregationMapping);

			verify(aggregator).withAggregationMapping(aggregationMapping);
		}

		@Test
		void shouldSetBatchSize() {
			runner.prepareAggregator(execution, aggregationMapping);

			verify(aggregator).withBatchSize(batchSize);
		}

		@Test
		void shouldReturnAggregator() {
			var result = runner.prepareAggregator(execution, aggregationMapping);

			assertThat(result).isEqualTo(aggregator);
		}
	}

	static class AggregationManagerRunnerImpl extends AggregationManagerRunner {

		public AggregationManagerRunnerImpl(AggregationManagerConfiguration config, TransformationProperties transformationProperties,
				TransformationService transformationService) {
			super(config, transformationProperties, transformationService);
		}

		@Override
		protected Aggregator createAggregator() {
			return aggregator;
		}
	}
}
