/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.aggregation.transformation;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;
import com.schibsted.spt.data.jslt.Expression;
import com.schibsted.spt.data.jslt.impl.DotExpression;
import com.schibsted.spt.data.jslt.impl.ExpressionImpl;
import com.schibsted.spt.data.jslt.impl.ObjectExpression;
import com.schibsted.spt.data.jslt.impl.PairExpression;
import com.schibsted.spt.data.jslt.impl.Scope;
import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.aggregation.transformation.AggregationMapping.FieldMapping;
import de.ozgcloud.apilib.vorgang.OzgCloudEingangHeaderTestFactory;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangEingangTestFactory;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangHeaderTestFactory;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangTestFactory;

class JSLTransformationServiceTest {

	@InjectMocks
	@Spy
	private JSLTransformationService service;
	@Mock
	private ObjectMapper objectMapper;
	@Mock
	private VorgangMapper vorgangMapper;

	@Nested
	class TestLoad {

		private final String identifier = LoremIpsum.getInstance().getWords(1);
		private final AggregationMapping aggregationMapping = AggregationMappingTestFactory.create();
		private final Map<String, String> mapping = FieldMappingTestFactory.createAsMap();

		@Mock
		private Expression expression;
		private Optional<DotExpression> identifierExpression = Optional.of(DotExpressionTestFactory.create());

		@Nested
		class TestOnMappingNotNull {
			@BeforeEach
			void mock() {
				doReturn(mapping).when(service).transformToMap(any());
				doReturn(expression).when(service).createExpression(any());
				doReturn(identifierExpression).when(service).transformToAbstractNode(any());
			}

			@Test
			void shouldTransformToMap() {
				load();

				verify(service).transformToMap(List.of(AggregationMappingTestFactory.MAPPING));
			}

			@Test
			void shouldCreateExpression() {
				load();

				verify(service).createExpression(mapping);
			}

			@Test
			void shouldReturnJSLTransformation() {
				var expectedTransformation = JSLTransformation.builder()
						.objectMapper(objectMapper)
						.vorgangMapper(vorgangMapper)
						.identifier(identifierExpression)
						.script(expression)
						.build();

				var transformation = load();

				assertThat(transformation).usingRecursiveComparison().isEqualTo(expectedTransformation);
			}

			private Transformation load() {
				return service.load(identifier, aggregationMapping);
			}
		}

		@Nested
		class TestOnMappingNull {
			@BeforeEach
			void mock() {
				doReturn(mapping).when(service).transformToMap(any());
				doReturn(expression).when(service).createExpression(any());
				doReturn(identifierExpression).when(service).transformToAbstractNode(any());
			}

			@Test
			void shouldTransformToMap() {
				load();

				verify(service).transformToMap(Collections.emptyList());
			}

			@Test
			void shouldCreateExpression() {
				load();

				verify(service).createExpression(mapping);
			}

			@Test
			void shouldReturnJSLTransformation() {
				var expectedTransformation = JSLTransformation.builder()
						.objectMapper(objectMapper)
						.vorgangMapper(vorgangMapper)
						.identifier(identifierExpression)
						.script(expression)
						.build();

				var transformation = load();

				assertThat(transformation).usingRecursiveComparison().isEqualTo(expectedTransformation);
			}

			private Transformation load() {
				return service.load(identifier, null);
			}
		}
	}

	@Nested
	class TestTransformToMap {

		private static final String SECOND_SOURCE = LoremIpsum.getInstance().getWords(1);
		private static final String SECOND_TARGET = LoremIpsum.getInstance().getWords(1);
		private final List<FieldMapping> mappings = List.of(
				FieldMappingTestFactory.create(),
				FieldMappingTestFactory.createBuilder().sourcePath(SECOND_SOURCE).targetPath(SECOND_TARGET).build());

		@Test
		void shouldTransformToMap() {
			var expectedMap = Map.of(FieldMappingTestFactory.TARGET_PATH, FieldMappingTestFactory.SOURCE_PATH, SECOND_TARGET, SECOND_SOURCE);

			var map = service.transformToMap(mappings);

			assertThat(map).containsExactlyInAnyOrderEntriesOf(expectedMap);
		}
	}

	@Nested
	class TestCreateExpression {

		@Test
		void shouldCallToPairExpression() {
			createExpression();

			verify(service).toPairExpression(Map.entry(FieldMappingTestFactory.TARGET_PATH, FieldMappingTestFactory.SOURCE_PATH));
		}

		@Test
		void shouldMapProperty() {

			var expression = service.createExpression(Map.of("target", "vorgangName"));

			var mapped = expression.apply(getVorgangTree());
			assertThat(mapped).hasToString("{\"target\":\"%s\"}".formatted(OzgCloudVorgangTestFactory.VORGANG_NAME));
		}

		private JsonNode getVorgangTree() {
			var mapper = new ObjectMapper().registerModules(new JavaTimeModule());
			mapper.configure(SerializationFeature.WRITE_DATES_AS_TIMESTAMPS, false);
			mapper.configure(SerializationFeature.WRITE_DATES_WITH_ZONE_ID, true);
			var vorgangTree = mapper.valueToTree(OzgCloudVorgangTestFactory.create());
			return vorgangTree;
		}

		@Nested
		class TestOnPairExpressionPresent {

			@Mock
			private PairExpression pairExpression;
			@Mock
			private ObjectExpression objectExpression;

			@BeforeEach
			void mock() {
				doReturn(Optional.of(pairExpression)).when(service).toPairExpression(any());
				doReturn(objectExpression).when(service).buildObjectExpression(any());
			}

			@Test
			void shouldBuildObjectExpression() {
				createExpression();

				verify(service).buildObjectExpression(new PairExpression[] { pairExpression });
			}

			@Test
			void shouldReturnObjectExpression() {
				var expression = createExpression();

				assertThat(expression).usingRecursiveComparison().isEqualTo(
						new ExpressionImpl(JSLTransformationService.EMPTY_TEMPLATES, JSLTransformationService.EMPTY_FUNCTIONS, objectExpression));
			}
		}

		@Nested
		class TestOnPairExpressionNotPresent {

			@Mock
			private ObjectExpression objectExpression;

			@BeforeEach
			void mock() {
				doReturn(Optional.empty()).when(service).toPairExpression(any());
				doReturn(objectExpression).when(service).buildObjectExpression();
			}

			@Test
			void shouldBuildObjectExpression() {
				createExpression();

				verify(service).buildObjectExpression();
			}

			@Test
			void shouldReturnObjectExpression() {
				var expression = createExpression();
				expression.apply(null);
				assertThat(expression).usingRecursiveComparison().isEqualTo(
						new ExpressionImpl(JSLTransformationService.EMPTY_TEMPLATES, JSLTransformationService.EMPTY_FUNCTIONS, objectExpression));
			}
		}

		private Expression createExpression() {
			return service.createExpression(FieldMappingTestFactory.createAsMap());
		}
	}

	@Nested
	class TestToPairExpression {

		@Test
		void shouldMapProperty() {
			var expression = service.transformToAbstractNode("vorgangName");

			var mapped = expression.get().apply(Scope.getRoot(2), getVorgangTree());
			assertThat(mapped).hasToString("\"%s\"".formatted(OzgCloudVorgangTestFactory.VORGANG_NAME));
		}

		@Test
		void shouldMapNestedProperty() {

			var expression = service.transformToAbstractNode("header.aktenzeichen");

			var mapped = expression.get().apply(Scope.getRoot(2), getVorgangTree());
			assertThat(mapped).hasToString("\"%s\"".formatted(OzgCloudVorgangHeaderTestFactory.AKTENZEICHEN));
		}

		@Test
		void shouldMapListElement() {

			var expression = service.transformToAbstractNode("eingangs[0]");

			var mapped = expression.get().apply(Scope.getRoot(2), getVorgangTree());

			assertThat(mapped).hasToString(getEingangTree().toString());
		}

		@Test
		void shouldMapListElementProperty() {

			var expression = service.transformToAbstractNode("eingangs[0].header.formEngineName");

			var mapped = expression.get().apply(Scope.getRoot(2), getVorgangTree());

			assertThat(mapped).hasToString("\"%s\"".formatted(OzgCloudEingangHeaderTestFactory.FORM_ENGINE_NAME));
		}

		private JsonNode getVorgangTree() {
			var mapper = new ObjectMapper().registerModules(new JavaTimeModule());
			mapper.configure(SerializationFeature.WRITE_DATES_AS_TIMESTAMPS, false);
			mapper.configure(SerializationFeature.WRITE_DATES_WITH_ZONE_ID, true);
			return mapper.valueToTree(OzgCloudVorgangTestFactory.create());
		}

		private JsonNode getEingangTree() {
			var mapper = new ObjectMapper().registerModules(new JavaTimeModule());
			mapper.configure(SerializationFeature.WRITE_DATES_AS_TIMESTAMPS, false);
			mapper.configure(SerializationFeature.WRITE_DATES_WITH_ZONE_ID, true);
			return mapper.valueToTree(OzgCloudVorgangEingangTestFactory.create());
		}

		@Nested
		class TestOnDotExpressionPresent {

			@Mock
			private DotExpression dotExpression;
			@Mock
			private PairExpression pairExpression;

			@BeforeEach
			void mock() {
				doReturn(Optional.of(dotExpression)).when(service).transformToAbstractNode(any());
				doReturn(pairExpression).when(service).buildPairExpression(any(), any());
			}

			@Test
			void shouldBuildPairExpression() {
				toPairExpression();

				verify(service).buildPairExpression(FieldMappingTestFactory.SOURCE_PATH,
						dotExpression);
			}

			@Test
			void shouldReturnPairExpression() {
				var returnedPairExpression = toPairExpression();

				assertThat(returnedPairExpression).contains(pairExpression);
			}
		}

		@Nested
		class TestOnDotExpressionNotPresent {

			@BeforeEach
			void mock() {
				doReturn(Optional.empty()).when(service).transformToAbstractNode(any());
			}

			@Test
			void shouldReturnEmpty() {
				var returnedPairExpression = toPairExpression();

				assertThat(returnedPairExpression).isEmpty();
			}
		}

		private Optional<PairExpression> toPairExpression() {
			return service.toPairExpression(Map.entry(FieldMappingTestFactory.SOURCE_PATH,
					FieldMappingTestFactory.TARGET_PATH));
		}
	}

	@Nested
	class TestTransformToAbstractNode {

		@Test
		void shouldReturnEmptyOnNullPath() {
			var dotExpression = service.transformToAbstractNode(null);

			assertThat(dotExpression).isEmpty();
		}

		@Test
		void shouldMapProperty() {
			var expression = service.transformToAbstractNode("vorgangName");

			var mapped = expression.get().apply(Scope.getRoot(2), getVorgangTree());
			assertThat(mapped).hasToString("\"%s\"".formatted(OzgCloudVorgangTestFactory.VORGANG_NAME));
		}

		@Test
		void shouldMapNestedProperty() {

			var expression = service.transformToAbstractNode("header.aktenzeichen");

			var mapped = expression.get().apply(Scope.getRoot(2), getVorgangTree());
			assertThat(mapped).hasToString("\"%s\"".formatted(OzgCloudVorgangHeaderTestFactory.AKTENZEICHEN));
		}

		@Test
		void shouldMapListElement() {

			var expression = service.transformToAbstractNode("eingangs[0]");

			var mapped = expression.get().apply(Scope.getRoot(2), getVorgangTree());

			assertThat(mapped).hasToString(getEingangTree().toString());
		}

		@Test
		void shouldMapListElementProperty() {

			var expression = service.transformToAbstractNode("eingangs[0].header.formEngineName");

			var mapped = expression.get().apply(Scope.getRoot(2), getVorgangTree());

			assertThat(mapped).hasToString("\"%s\"".formatted(OzgCloudEingangHeaderTestFactory.FORM_ENGINE_NAME));
		}

		private JsonNode getVorgangTree() {
			var mapper = new ObjectMapper().registerModules(new JavaTimeModule());
			mapper.configure(SerializationFeature.WRITE_DATES_AS_TIMESTAMPS, false);
			mapper.configure(SerializationFeature.WRITE_DATES_WITH_ZONE_ID, true);
			return mapper.valueToTree(OzgCloudVorgangTestFactory.create());
		}

		private JsonNode getEingangTree() {
			var mapper = new ObjectMapper().registerModules(new JavaTimeModule());
			mapper.configure(SerializationFeature.WRITE_DATES_AS_TIMESTAMPS, false);
			mapper.configure(SerializationFeature.WRITE_DATES_WITH_ZONE_ID, true);
			return mapper.valueToTree(OzgCloudVorgangEingangTestFactory.create());
		}
	}
}
